"""
porter2.py from https://bitbucket.org/mchaput/stemming

An implementation of the Porter2 stemming algorithm.
See http://snowball.tartarus.org/algorithms/english/stemmer.html

Adapted from pyporter2 by Michael Dirolf.

This algorithm is more correct but (at least in this implementation)
several times slower than the original porter algorithm as implemented
in stemming.porter.
"""

import re
 
r_exp = re.compile(r"[^aeiouy]*[aeiouy]+[^aeiouy](\w*)")
ewss_exp1 = re.compile(r"^[aeiouy][^aeiouy]$")
ewss_exp2 = re.compile(r".*[^aeiouy][aeiouy][^aeiouywxY]$")
ccy_exp = re.compile(r"([aeiouy])y")
s1a_exp = re.compile(r"[aeiouy].")
s1b_exp = re.compile(r"[aeiouy]")

def get_r1(word):
    # exceptional forms
    if word.startswith('gener') or word.startswith('arsen'):
        return 5
    if word.startswith('commun'):
        return 6
 
    # normal form
    match = r_exp.match(word)
    if match:
        return match.start(1)
    return len(word)
 
def get_r2(word):
    match = r_exp.match(word, get_r1(word))
    if match:
        return match.start(1)
    return len(word)
 
def ends_with_short_syllable(word):
    if len(word) == 2:
        if ewss_exp1.match(word):
            return True
    if ewss_exp2.match(word):
        return True
    return False
 
def is_short_word(word):
    if ends_with_short_syllable(word):
        if get_r1(word) == len(word):
            return True
    return False
 
def remove_initial_apostrophe(word):
    if word.startswith("'"):
        return word[1:]
    return word
 
def capitalize_consonant_ys(word):
    if word.startswith('y'):
        word = 'Y' + word[1:]
    return ccy_exp.sub('\g<1>Y', word)
 
def step_0(word):
    if word.endswith("'s'"):
        return word[:-3]
    if word.endswith("'s"):
        return word[:-2]
    if word.endswith("'"):
        return word[:-1]
    return word
 
def step_1a(word):
    if word.endswith('sses'):
        return word[:-4] + 'ss'
    if word.endswith('ied') or word.endswith('ies'):
        if len(word) > 4:
            return word[:-3] + 'i'
        else:
            return word[:-3] + 'ie'
    if word.endswith('us') or word.endswith('ss'):
        return word
    if word.endswith('s'):
        preceding = word[:-1]
        if s1a_exp.search(preceding):
            return preceding
        return word
    return word

doubles = ('bb', 'dd', 'ff', 'gg', 'mm', 'nn', 'pp', 'rr', 'tt')
def ends_with_double(word):
    for double in doubles:
        if word.endswith(double):
            return True
    return False
def step_1b_helper(word):
    if word.endswith('at') or word.endswith('bl') or word.endswith('iz'):
        return word + 'e'
    if ends_with_double(word):
        return word[:-1]
    if is_short_word(word):
        return word + 'e'
    return word
s1b_suffixes = ('ed', 'edly', 'ing', 'ingly')

def step_1b(word, r1):
    if word.endswith('eedly'):
        if len(word) - 5 >= r1:
            return word[:-3]
        return word
    if word.endswith('eed'):
        if len(word) - 3 >= r1:
            return word[:-1]
        return word
 
    for suffix in s1b_suffixes:
        if word.endswith(suffix):
            preceding = word[:-len(suffix)]
            if s1b_exp.search(preceding):
                return step_1b_helper(preceding)
            return word
 
    return word
 
def step_1c(word):
    if word.endswith('y') or word.endswith('Y') and len(word) > 1:
        if word[-2] not in 'aeiouy':
            if len(word) > 2:
                return word[:-1] + 'i'
    return word

def step_2_helper(word, r1, end, repl, prev):
        if word.endswith(end):
            if len(word) - len(end) >= r1:
                if prev == []:
                    return word[:-len(end)] + repl
                for p in prev:
                    if word[:-len(end)].endswith(p):
                        return word[:-len(end)] + repl
            return word
        return None
s2_triples = (('ization', 'ize', []),
               ('ational', 'ate', []),
               ('fulness', 'ful', []),
               ('ousness', 'ous', []),
               ('iveness', 'ive', []),
               ('tional', 'tion', []),
               ('biliti', 'ble', []),
               ('lessli', 'less', []),
               ('entli', 'ent', []),
               ('ation', 'ate', []),
               ('alism', 'al', []),
               ('aliti', 'al', []),
               ('ousli', 'ous', []),
               ('iviti', 'ive', []),
               ('fulli', 'ful', []),
               ('enci', 'ence', []),
               ('anci', 'ance', []),
               ('abli', 'able', []),
               ('izer', 'ize', []),
               ('ator', 'ate', []),
               ('alli', 'al', []),
               ('bli', 'ble', []),
               ('ogi', 'og', ['l']),
               ('li', '', ['c', 'd', 'e', 'g', 'h', 'k', 'm', 'n', 'r', 't']))

def step_2(word, r1):
    for trip in s2_triples:
        attempt = step_2_helper(word, r1, trip[0], trip[1], trip[2])
        if attempt:
            return attempt
    return word

def step_3_helper(word, r1, r2, end, repl, r2_necessary):
    if word.endswith(end):
        if len(word) - len(end) >= r1:
            if not r2_necessary:
                return word[:-len(end)] + repl
            else:
                if len(word) - len(end) >= r2:
                    return word[:-len(end)] + repl
        return word
    return None
s3_triples = (('ational', 'ate', False),
               ('tional', 'tion', False),
               ('alize', 'al', False),
               ('icate', 'ic', False),
               ('iciti', 'ic', False),
               ('ative', '', True),
               ('ical', 'ic', False),
               ('ness', '', False),
               ('ful', '', False))
def step_3(word, r1, r2):
    for trip in s3_triples:
        attempt = step_3_helper(word, r1, r2, trip[0], trip[1], trip[2])
        if attempt:
            return attempt
    return word

s4_delete_list = ('al', 'ance', 'ence', 'er', 'ic', 'able', 'ible', 'ant', 'ement',
                  'ment', 'ent', 'ism', 'ate', 'iti', 'ous', 'ive', 'ize')

def step_4(word, r2):
    for end in s4_delete_list:
        if word.endswith(end):
            if len(word) - len(end) >= r2:
                return word[:-len(end)]
            return word
 
    if word.endswith('sion') or word.endswith('tion'):
        if len(word) - 3 >= r2:
            return word[:-3]
 
    return word
 
def step_5(word, r1, r2):
    if word.endswith('l'):
        if len(word) - 1 >= r2 and word[-2] == 'l':
            return word[:-1]
        return word
 
    if word.endswith('e'):
        if len(word) - 1 >= r2:
            return word[:-1]
        if len(word) - 1 >= r1 and not ends_with_short_syllable(word[:-1]):
            return word[:-1]
 
    return word
 
def normalize_ys(word):
    return word.replace('Y', 'y')
 
exceptional_forms = {'skis': 'ski',
                    'skies': 'sky',
                    'dying': 'die',
                    'lying': 'lie',
                    'tying': 'tie',
                    'idly': 'idl',
                    'gently': 'gentl',
                    'ugly': 'ugli',
                    'early': 'earli',
                    'only': 'onli',
                    'singly': 'singl',
                    'sky': 'sky',
                    'news': 'news',
                    'howe': 'howe',
                    'atlas': 'atlas',
                    'cosmos': 'cosmos',
                    'bias': 'bias',
                    'andes': 'andes'}
 
exceptional_early_exit_post_1a = frozenset(['inning', 'outing', 'canning', 'herring',
                                            'earring', 'proceed', 'exceed', 'succeed'])
 

def stem(word):
    if len(word) <= 2:
        return word
    word = remove_initial_apostrophe(word)

    # handle some exceptional forms
    if word in exceptional_forms:
        return exceptional_forms[word]

    word = capitalize_consonant_ys(word)
    r1 = get_r1(word)
    r2 = get_r2(word)
    word = step_0(word)
    word = step_1a(word)

    # handle some more exceptional forms
    if word in exceptional_early_exit_post_1a:
        return word

    word = step_1b(word, r1)
    word = step_1c(word)
    word = step_2(word, r1)
    word = step_3(word, r1, r2)
    word = step_4(word, r2)
    word = step_5(word, r1, r2)
    word = normalize_ys(word)

    return word

if __name__ == "__main__":
    assert stem("bill's") == "bill"
    assert stem("y's") == "y"



import sys, os, threading, queue
sys.path.append('.')
os.chdir('..')
import normalize
from singleton import db

num_workers = 64

in_q = queue.Queue()
out_q = queue.Queue()
class Worker(threading.Thread):
  def run(self):
    while True:
      uid, url = in_q.get()
      if uid is None:
        out_q.put((None, None, None))
        return
      new_url = normalize.dereference(url)
      if url != new_url:
        out_q.put((uid, url, new_url))

workers = []
for i in range(num_workers):
  workers.append(Worker())
  workers[-1].setDaemon(True)
  workers[-1].start()

c = db.cursor()
c.execute("""select item_uid, item_link
from fm_items
where item_rating>0
order by item_uid""")
list(map(in_q.put, c))
list(map(in_q.put, [(None, None)] * num_workers))


while True:
  uid, url, new_url = out_q.get()
  if uid is None and url is None and new_url is None:
    num_workers -= 1
    if num_workers == 0:
      db.commit()
      sys.exit(0)
    continue
  print(uid, url)
  print('\t==>', new_url)
  c.execute('update fm_items set item_link=? where item_uid=?',
            [new_url, uid])



import time, bisect
from . import param, dbop

urls = []

with dbop.db() as db:
  before = time.time()
  print('loading existing URLs', file=param.activity)
  c = db.execute("select item_link from fm_items")
  
  urls = set(row[0] for row in c)
  print('URL loading', len(urls), 'URLs done in', time.time() - before,
        file=param.activity)
  before = time.time()
  urls = list(urls)
  urls.sort()
  print('sort done in', time.time() - before, file=param.activity)

def add(url):
  i = bisect.bisect_left(urls, url)
  # do not insert a duplicate
  if i < len(urls) and urls[i] == url:
    return
  urls.insert(i, url)

def rename(old, new):
  i = bisect.bisect_left(urls, old)
  if i < len(urls) and urls[i] == old:
    urls[i] = new
    urls.sort()

def exists(url):
  i = bisect.bisect_left(urls, url)
  # match if there is an existing superstring of the URL
  return i < len(urls) and url in urls[i]



# handle the various type of FilteringRules
from __future__ import print_function
import time, re, textwrap, requests, html5lib
from . import normalize, param, util, urldedup

rules = []
feed_rules = {}
loaded = False

def evaluate_rules(item, feed, feed_uid, exempt):
  # item.__getitem__ is expensive so cache the most frequently used members
  cached = dict()
  for target in ('title', 'content'):
    for suffix in ('', '_words', '_words_exact', '_lc'):
      cached[target + suffix] = item[target + suffix]
  # now loop through the applicable rules
  for rule in rules * (not exempt) + feed_rules.get(feed_uid, list()):
    try:
      if rule.test(item, feed, feed_uid, cached):
        return True, rule
    except:
      util.print_stack(['f'])
  return False, None

class Rule:
  registry = dict()
  def __init__(self, uid, expires):
    assert uid not in self.registry
    self.registry[uid] = self
    self.uid = uid
    self.expires = expires
  def __str__(self):
    return '<Rule %s>' % self.uid
  def __repr__(self):
    return self.__str__()
  def check_expires(self):
    return self.expires and time.time() > self.expires
  def highlight_title(self, html):
    return html
  def highlight_content(self, html):
    return html

class KeywordRule(Rule):
  def __init__(self, uid, expires, rule, rtype):
    Rule.__init__(self, uid, expires)
    self.target, self.match = rtype.split('_', 1)
    assert self.target in ['title', 'content']
    if self.match in ['exactword']:
      self.rule = normalize.get_words(rule)
    elif self.match in ['word', 'all']:
      self.rule = normalize.stem(normalize.get_words(rule))
    else:
      self.rule = rule
  def __str__(self):
    return '<KeywordRule %s %s %s %s>' % (self.uid, self.target, self.match,
                                          self.rule)
  def test(self, item, feed, feed_uid, cached):
    if self.check_expires():
      return False
    if self.match in ['word', 'all']:
      suffix = '_words'
    elif self.match == 'exactword':
      suffix = '_words_exact'
    elif self.match == 'phrase_lc':
      suffix = '_lc'
    else:
      suffix = ''
    target = cached[self.target + suffix]
    if self.match in ('word', 'exactword'):
      return bool(target.intersection(self.rule))
    elif self.match == 'exactword':
      return bool(target.intersection(self.rule))
    elif self.match == 'all':
      return bool(target.issuperset(self.rule))
    else:
      return self.rule in target
  def highlight(self, html):
    if type(self.rule) in [str, str]:
      return normalize.replace_first(
        html, self.rule, '<span class="filter-highlight">', '</span>')
    else:
      for word in self.rule:
        html = normalize.replace_first(
          html, word, '<span class="filter-highlight">', '</span>')
      return html
  def highlight_title(self, html):
    if self.target == 'content' \
           and self.uid > 0 and -self.uid in self.registry:
      return  self.highlight(html)
    if self.target == 'title':
      return  self.highlight(html)
    return html
  def highlight_content(self, html):
    if self.target == 'content':
      return  self.highlight(html)
    return html

class TagRule(Rule):
  def __init__(self, uid, expires, rule):
    Rule.__init__(self, uid, expires)
    self.rule = normalize.lower(rule)
  def __str__(self):
    return '<TagRule %s %s>' % (self.uid, self.rule)
  def test(self, item, feed, feed_uid, cached):
    if self.check_expires():
      return False
    for tag in item['item_tags']:
      if self.rule == tag:
        return True
    return False
  def highlight_content(self, html):
    return '%s<br><p>Filtered for tag <span class="item tag highlighted">%s</span></p>' \
           % (html, self.rule)

class AuthorRule(Rule):
  def __init__(self, uid, expires, rule):
    Rule.__init__(self, uid, expires)
    self.rule = rule
  def __str__(self):
    return '<AuthorRule %s %s>' % (self.uid, self.rule)
  def test(self, item, feed, feed_uid, cached):
    if self.check_expires():
      return False
    return self.rule == normalize.lower(item['author'])
  def highlight_content(self, html):
    return '%s<br><p>Filtered for author <span class="author tag highlighted">%s</span></p>' \
           % (html, self.rule)

########################################################################
# functions used inside Python rules
def link_already(url):
  print('checking for deja-vu for', url, end=' ', file=param.activity)
  return urldedup.exists(url)

def link_extract(link_text, content):
  """Extract first link after link_text"""
  h = html5lib.parse(content, namespaceHTMLElements=False)
  candidates = h.findall(".//a[.='%s']" % link_text)
  if not candidates:
    return 'NOT MATCHED'
  try:
    return candidates[0].attrib['href']
  except:
    return 'NOT MATCHED'

def dereference_content(url):
  try:
    r = requests.get(url, timeout=param.http_timeout)
    return r.content
  except:
    return ''

# shades of LISP...
def curry(fn, obj):
  return lambda *args: fn(obj, *args)

# obj can be a string, list or dictionary
def any(obj, *words):
  for w in words:
    if w in obj:
      return True
  return False

def union_any(obj_list, *words):
  for w in words:
    for obj in obj_list:
      if w in obj:
        return True
  return False

########################################################################

rule_comment_re = re.compile('^#.*$', re.MULTILINE)
def normalize_rule(rule):
  """allow embedded CR/LF and comments to make for more readable rules"""
  return rule_comment_re.sub('', rule).replace(
    '\n', ' ').replace('\r', ' ').strip()

wrapper = textwrap.TextWrapper(width=80, break_long_words=False)
# XXX this relies on texwrap implementation details to prevent wrapping on
# XXX hyphens and em-dashes, only on spaces
wrapper.wordsep_re = re.compile(r'(\s+)')
def rule_lines(rule):
  "Find how many lines are needed for the rule in a word-wrapped <textarea>"
  if not rule:
    return 4
  lines = 0
  for line in rule.splitlines():
    if line.strip():
      lines += len(wrapper.wrap(line))
    else:
      lines += 1
  return lines

class PythonRule(Rule):
  def __init__(self, uid, expires, rule):
    Rule.__init__(self, uid, expires)
    self.rule = rule
    rule = normalize_rule(rule)
    self.code = compile(rule, 'rule' + repr(uid), 'eval')
  def __str__(self):
    return '<PythonRule %s %s>' % (self.uid, normalize_rule(self.rule))
  def test(self, item, feed, feed_uid, cached):
    if self.check_expires():
      return False
    filter_dict = dict()
    for key in feed.feed:
      try:
        filter_dict['feed_' + key] = feed.feed[key]
      except KeyError:
        pass
    filter_dict.update(item)
    # for backward compatibility, see normalize.py
    filter_dict['category'] = filter_dict['item_tags']
    # used to filter echos from sites like Digg
    filter_dict['link_already'] = link_already
    filter_dict['link_extract'] = link_extract
    filter_dict['dereference_content'] = dereference_content
    # convenient shortcut functions
    filter_dict['title_any_words'] = curry(any, item['title_words'])
    filter_dict['content_any_words'] = curry(any, item['content_words'])
    filter_dict['union_any_words'] = curry(
      union_any, [item['title_words'], item['content_words']])
    filter_dict['title_any'] = curry(any, item['title'])
    filter_dict['content_any'] = curry(any, item['content'])
    filter_dict['union_any'] = curry(
      union_any, [item['title'], item['content']])
    filter_dict['title_any_lc'] = curry(any, item['title_lc'])
    filter_dict['content_any_lc'] = curry(any, item['content_lc'])
    filter_dict['union_any_lc'] = curry(
      union_any, [item['title_lc'], item['content_lc']])
    return bool(eval(self.code, filter_dict))
  def highlight_title(self, html):
    return html
  def highlight_content(self, html):
    return html + '<br><p>Filtered by Python rule %d</p>' % self.uid
    
def load_rules(c):
  global loaded, rules, feed_rules
  if loaded: return
  rules = []
  feed_rules = dict()
  try:
    try:
      for uid, rtype, rule, feed_uid, expires in \
          c.execute("""select rule_uid, rule_type, rule_text, rule_feed_uid,
          strftime('%s', rule_expires)
          from fm_rules
          where rule_expires is null or rule_expires > julianday('now')"""):
        if expires: expires = int(expires)
        if feed_uid:
          container = feed_rules.setdefault(feed_uid, list())
        else:
          container = rules
        if rtype == 'python':
          rule = PythonRule(uid, expires, rule)
          container.append(rule)
        elif rtype == 'tag':
          rule = TagRule(uid, expires, rule)
          container.append(rule)
        elif rtype == 'author':
          rule = AuthorRule(uid, expires, rule)
          container.append(rule)
        elif rtype.startswith('union_'):
          # XXX this convention of adding a second rule object with UID -uid
          # XXX is a ugly hack
          container.append(KeywordRule(
            -uid, expires, rule, rtype.replace('union_', 'title_')))
          container.append(KeywordRule(
            uid, expires, rule, rtype.replace('union_', 'content_')))
        else:
          container.append(KeywordRule(uid, expires, rule, rtype))
      for feed_uid, rule in \
          c.execute("""select feed_uid, feed_filter from fm_feeds
          where feed_filter is not null"""):
        rule = PythonRule('feed_%d' % feed_uid, None, rule)
        feed_rules.setdefault(feed_uid, list()).append(rule)
    except:
      util.print_stack()
  finally:
    loaded = True

def invalidate():
  """Invalidate the rule cache to force reloading from the database"""
  # break cyclic references
  Rule.registry.clear()
  global loaded
  loaded = False

def update_rule(db, c, uid, expires, text, delete):
  if expires == 'never':
    expires = 'NULL'
  else:
    expires = "julianday('%s')" % expires
  # check syntax
  compile(normalize_rule(text), 'web form', 'eval')
  if uid == 'new':
    c.execute("insert into fm_rules (rule_expires, rule_text) values (?, ?)",
              [expires, text])
  elif delete == 'on':
    c.execute("delete from fm_rules where rule_uid=?", [uid])
  else:
    c.execute("""update fm_rules set rule_expires=?, rule_text=?
    where rule_uid=?""", [expires, text, uid])
  db.commit()
  invalidate()

def add_kw_rule(db, c, kw=None, item_uid=None, match='word', target='title',
                feed_only=False, retroactive=False, stem=None, **kwargs):
  feed_only = bool(feed_only)
  retroactive = bool(retroactive)

  if feed_only:
    item_uid = int(item_uid)
  else:
    item_uid = None

  if match == 'word':
    kw = stem
  if not kw: return
  if match in ['author', 'tag', 'phrase_lc']:
    words = [normalize.lower(kw)]
  elif match in {'word', 'exactword'}:
    words = normalize.get_words(kw)
  elif match == 'all':
    words = [' '.join(normalize.get_words(kw))]
  elif match == 'phrase':
      words = [kw]
  else:
    return
  
  if match in ['author', 'tag']:
    rule_type = match
  else:
    rule_type = target + '_' + match

  for word in words:
    print('ADD_KW_RULES', rule_type, item_uid, word, file=param.log)
    c.execute("""insert into fm_rules (rule_type, rule_feed_uid, rule_text)
    values (?, (select item_feed_uid from fm_items where item_uid=?), ?)""",
              [rule_type, item_uid, word]);
  invalidate()

def del_kw_rule(db, c, rule_uid=None, **kwargs):
  c.execute("""update fm_items
  set item_rating=0, item_rule_uid=NULL
  where item_rule_uid=? and item_content!=''""", [rule_uid])
  c.execute('delete from fm_rules where rule_uid=?', [rule_uid])
  invalidate()

def exempt_feed_retroactive(db, c, feed_uid, **kwargs):
  """Retroactively unfilter a feed that is exempted from filtering"""
  c.execute("""update fm_items
  set item_rating=0, item_rule_uid=NULL
  where item_feed_uid=? and item_content!='' and exists (
    select rule_uid from fm_rules
    where rule_feed_uid is null and item_rule_uid=rule_uid
  )""", [feed_uid])

########################################################################
# stats
def stats(c):
  return c.execute("""select rule_uid, rule_type, rule_text,
    coalesce(rule_feed_uid, -1), feed_title,
    sum(case when item_created > julianday('now')-7 then 1 else 0 end) last_7,
    sum(case when item_created < julianday('now')-7 then 1 else 0 end) prev_7,
    min(case when item_created > julianday('now')-7
             then item_uid else 2000000000 end),
    max(case when item_created > julianday('now')-7
             then item_uid else 0 end)
  from fm_rules
  join fm_items on item_rule_uid=rule_uid
  left join fm_feeds on rule_feed_uid=feed_uid
  where item_created > julianday('now') -14
  group by 1, 2, 3, 4, 5
  order by 6 desc
  limit 100""")



# essential sanity checks, do not start temboz if these fail
import sys, os.path, pprint, feedparser, bleach
import tembozapp.normalize as normalize

bugfeed = (os.path.dirname(__file__) or '.') + os.sep + 'bugfeed'
#print(__file__, bugfeed)
calmatters = open(bugfeed + os.sep + 'calmatters')
text = calmatters.read()
calmatters.close()
#import pdb
#pdb.set_trace()

f = feedparser.parse(text)
for i in f.entries:
  normalize.normalize(i, f, False)
  #pprint.pprint(i)
  assert '<script>' not in repr(i), \
    'Refusing to start Temboz because feedparser is not sanitizing properly'

assert bleach.__version__ >= '5', \
  'Temboz now requires Bleach 5 or later, which made breaking API changes'



#!/usr/local/bin/python
import sys, time, os, threading, unittest, pprint
import feedparser

class TestCase(unittest.TestCase):
  def setUp(self):
    pass

  def test100_sanitize(self):
    """Test feedparser HTML sanitizer"""
    import sanity

def suite():
  suite = unittest.makeSuite(TestCase, 'test')
  return suite

if __name__ == '__main__':
  unittest.main()



def fts5_term(term):
  """Convert a Google search like query into a FTS5 query, i.e. to not
  trigger the column filter misfeature. We don't try to make invalid queries
  work, just make queries that would be interpreted as column filters or the
  like work.

  EBNF:
  query = word
  word = [not space]+
  query = phrase
  phrase = '"' [^"]* '"'
  query = "(" query ")"
  query = query "AND" query
  query = query "OR" query
  query = "NOT" query
  """
  in_q = False
  implicit_q = False
  out = []
  expect = None
  pending = []
  WS = 'WORDSEP'
  #print('-'*72)
  #print(f'Parsing {term}')
  for c in term:
    #print(f'c={c} in_q={in_q} implicit_q={implicit_q} out={out} expect={expect} pending={pending}')
    if c == "'": # SQL injection guard
      if pending:
        if not implicit_q:
          out.append('"')
          implicit_q = True
        out.extend(pending)
        pending = None
      out.append("''")
      continue
    if expect:
      assert not in_q
      if expect == WS and c not in ' \t\n("' or expect != WS and c != expect:
        out.append('"')
        implicit_q = True
        out.extend(pending)
        expect = None
        pending = None
      elif expect == WS:
        out.extend(pending)
        expect = None
        pending = None
      else:
        pending.append(c)
        expect = {'N': 'D', 'D': WS, 'R': WS, 'O': 'T', 'T': WS}[expect]
        continue
    if c == '"':
      if implicit_q:
        implicit_q = False
        in_q = True
      else:
        in_q = not in_q
        out.append(c)
    elif not in_q and not implicit_q and c in 'AON':
      expect = {'A': 'N', 'O': 'R', 'N': 'O'}[c]
      pending = [c]
    elif c in ' \t\n()':
      if implicit_q:
        out.append('"')
        implicit_q = False
        in_q = False
      out.append(c)
    else:
      if in_q:
        out.append(c)
      elif implicit_q:
        out.append(c)
      else:
        in_q = True
        implicit_q = True
        out.append('"')
        out.append(c)
  if pending:
    out.extend(pending)
  if in_q or implicit_q:
    out.append('"')
  return ''.join(out)

if __name__ == '__main__':
  fts5_term('foo AND bar')
  for input, expected in [
      ('foo', '"foo"'),
      ('foo bar', '"foo" "bar"'),
      ('"foo"', '"foo"'),
      ('"foo bar"', '"foo bar"'),
      ('foo AND bar', '"foo" AND "bar"'),
      ('(foo AND bar) OR baz', '("foo" AND "bar") OR "baz"'),
      ('foo AN bar', '"foo" "AN" "bar"'),
      ('"foo AN bar"', '"foo AN bar"'),
      ('ACME', '"ACME"'),
      ('Acme', '"Acme"'),
      ('OpenBSD', '"OpenBSD"'),
      ('ANAN', '"ANAN"'),
      ('ANAND', '"ANAND"'),
      ('AN"D"', '"AND"'),
      ('NOTAM', '"NOTAM"'),
      ('ANDROS', '"ANDROS"'),
  ]:
    out = fts5_term(input)
    assert out == expected, \
      'input=%r expected=%r got=%r' % (input, expected, out)



########################################################################
#
# Parameter file for Temboz
#
########################################################################

# number of RSS feeds fetched in parallel 
feed_concurrency = 20

# Maximum number of articles shown
overload_threshold = 200

# feed polling interval in seconds
refresh_interval = 3600

# half-life of articles' contribution towards the SNR
decay = 90

# Whether "catch-up" links require user confirmation (default is yes)
catch_up_confirm = True
hard_purge_confirm = True

# automatic backups
# stream compression utility to use for backups
backup_compressor = ('gzip -9c', '.gz')
#backup_compressor = ('bzip2 -9c', '.bz2)
# number of daily backups to keep
daily_backups = 7
# at what time should the backup be made (default: between 3 and 4 AM)
backup_hour = 4

# garbage collection - articles flagged as "uninteresting" will have their
# content automatically dumped after this interval (but not their title or
# permalink) to make room. If this parameter is set to None or False, this
# garbage-collection will not occur
garbage_contents = 7
# garbage_contents = None

# after a longer period of time, the articles themselves are purged, assuming
# they are no longer in the feed files (otherwise they would reappear the next
# time the feed is loaded)
# Note: this needs to be set much higher than the healf life for SNR
garbage_items = 180
# garbage_items = None

# URL to use as the User-Agent when downloading feeds
temboz_url = 'https://www.temboz.com/'
# user agent shown when fetching the feeds
user_agent = 'Temboz (%s)' % temboz_url
def default_user_agent():
  return user_agent
import requests
requests.utils.default_user_agent = default_user_agent

# page unauthenticated users should see
# the most common case is people checking the referrer logs on their web server
unauth_page = temboz_url

# dictionary of login/password
try:
  from private import auth_dict
except:
  auth_dict = {'majid': 'sopo'}

# maximum number of errors, after this threshold is reached,
# the feed is automatically suspended. -1 to unlimit
max_errors = 1000

#debug = True
debug = False
#debug_sql = True
debug_sql = False
#profile = False

# logging
import sys, os
log_filename = 'error.log'
if '--daemon' in sys.argv:
  # if you modify mode and buffer size, see also update.py:cleanup
  # for the code that rotates this file daily
  log = open(log_filename, 'a', 0)
  os.dup2(log.fileno(), 1)
  os.dup2(log.fileno(), 2)
  activity = open('activity.log', 'a')
else:
  log = sys.stderr
  activity = sys.stderr
# redirect stout and stderr to the log file

# default timeout for HTTP requests in seconds
http_timeout = 60.0

# These settings are managed in the database and will ultimately supersede
# param.py
settings = {}



#!/usr/bin/env python3
import sys, time, os, threading, unittest, pprint, subprocess
import feedparser
import tembozapp.feedfix, tembozapp.normalize, tembozapp.filters

class TestCase(unittest.TestCase):
  def setUp(self):
    pass

  def test100_autodiscovery(self):
    """Test """
    import tembozapp.autodiscovery
    feed_url = tembozapp.autodiscovery.find('https://blog.majid.info/')
    assert feed_url == 'https://blog.majid.info/index.xml'
    feed_url = tembozapp.autodiscovery.find('https://greenwald.substack.com')
    assert feed_url == 'https://greenwald.substack.com/feed/'
    feed_url = tembozapp.autodiscovery.find('https://www.calnewport.com/blog/')
    assert feed_url == 'https://www.calnewport.com/blog/feed'

  def test101_feedparser_exceptions(self):
    for fn in os.listdir('bugfeed'):
      f = open('bugfeed/' + fn, 'r')
      d = f.read()
      f.close()
      feedparser.parse(d)

  def test102_title_xss(self):
    f = open('bugfeed/boingboing', 'r')
    d = f.read()
    f.close()
    f = feedparser.parse(d)
    i = [i for i in f.entries if '>' in i.title][0]
    tembozapp.normalize.normalize(i, f, False)
    assert '>' not in i.title

  def test103_highlight(self):
    uid = 0
    for (kind, pat, s) in [
        ('title_word', 'cloud', "Sopra Steria gets £££££££s to manage cops' Oracle e-Biz suite in Oracle's cloud in Cleveland, UK"),
        ('title_word', 'gb', "Sopra Steria gets £££££££s to manage cops' Oracle e-Biz suite in Oracle's cloud in Cleveland, UK"),
    ]:
      uid += 1
      rule = tembozapp.filters.KeywordRule(
        uid, None, pat, kind)
      before = time.time()
      s2 = rule.highlight_title(s)
      after = time.time()
      assert after - before < 1.0
      assert s2 != s
      assert '<span' in s2
    
def suite():
  suite = unittest.makeSuite(TestCase, 'test')
  return suite

if __name__ == '__main__':
  unittest.main()



from __future__ import print_function
import sys, traceback
import tembozapp.param as param

# Utility functions for debugging
# we have to be extra defensive in order not to erase the original exception
# with one generated in this function
# the argument 'black_list' is a list of variable names that we don't want
# to see when we display the local variable dictionary.
def print_stack(black_list=[]):
  e = sys.exc_info()
  print('#' * 10, 'BEGIN', '#' * 60, file=param.log)
  print(str(e[0]) +':', e[1], file=param.log)
  if e[1] != None:
    traceback.print_exc(None, param.log)
    t = e[2]
    # this should not be necessary as the tracback should have at least two
    # elements, one for print_stack and one for the caller function
    if t != None and t.tb_next != None:
      while t.tb_next.tb_next != None:
        t = t.tb_next
    if t != None:
      print('-' * 10, 'local variables:', '-' * 50, file=param.log)
      for var_name, var_value in list(t.tb_frame.f_locals.items()):
        if var_name not in black_list:
          print(var_name, ':', repr(var_value), file=param.log)
    del t
  # to help the garbage collector
  del e
  print('#' * 10, 'END', '#' * 63, file=param.log)
